# -*- coding: utf-8 -*-
"""This module contains classes handling different treatments of ozone."""

import abc
import logging
from scipy.interpolate import interp1d

from konrad.component import Component
from konrad.utils import ozone_profile_rcemip, refined_pgrid

__all__ = [
    'Ozone',
    'OzonePressure',
    'OzoneHeight',
    'OzoneNormedPressure',
]

logger = logging.getLogger(__name__)


class Ozone(Component, metaclass=abc.ABCMeta):
    """Base class to define abstract methods for ozone treatments."""

    def __init__(self):
        """
        Parameters:
            initial_ozone (ndarray): initial ozone vmr profile
        """
        self['initial_ozone'] = (('plev',), None)

    @abc.abstractmethod
    def __call__(self, atmosphere, convection, timestep, zenith):
        """Updates the ozone profile within the atmosphere class.

        Parameters:
            atmosphere (konrad.atmosphere): atmosphere model containing ozone
                concentration profile, height, temperature, pressure and half
                pressure levels at the current timestep
            convection (konrad.convection): convection scheme
            timestep (float): timestep of run [days]
            zenith (float): solar zenith angle,
                angle of the Sun to the vertical [degrees]
        """


class OzonePressure(Ozone):
    """Ozone fixed with pressure, no adjustment needed."""
    def __call__(self, **kwargs):
        return


class OzoneHeight(Ozone):
    """Ozone fixed with height."""
    def __init__(self):
        self._f = None

    def __call__(self, atmosphere, **kwargs):
        if self._f is None:
            self._f = interp1d(
                atmosphere['z'][0, :],
                atmosphere['O3'],
                fill_value='extrapolate',
            )

        atmosphere['O3'] = (('time', 'plev'), self._f(atmosphere['z'][0, :]))


class OzoneNormedPressure(Ozone):
    """Ozone shifts with the normalisation level (chosen to be the convective
    top)."""
    def __init__(self, norm_level=None):
        """
        Parameters:
            norm_level (float): pressure for the normalisation
                normally chosen as the convective top pressure at the start of
                the simulation [Pa]
        """
        self.norm_level = norm_level
        self._f = None

    def __call__(self, atmosphere, convection, **kwargs):
        if self.norm_level is None:
            self.norm_level = convection.get('convective_top_plev')[0]
            # TODO: what if there is no convective top

        if self._f is None:
            self._f = interp1d(
                atmosphere['plev'] / self.norm_level,
                atmosphere['O3'][0, :],
                fill_value='extrapolate',
            )

        norm_new = convection.get('convective_top_plev')[0]

        atmosphere['O3'] = (
            ('time', 'plev'),
            self._f(atmosphere['plev'] / norm_new).reshape(1, -1)
        )



# -*- coding: utf-8 -*-
"""Common utility functions.
"""
import copy
import logging
from datetime import timedelta
from numbers import Number

import numpy as np
import typhon as ty
from netCDF4 import Dataset

from konrad import constants


__all__ = [
    'append_description',
    'append_timestep_netcdf',
    'return_if_type',
    'phlev_from_plev',
    'dz_from_z',
    'refined_pgrid',
    'get_pressure_grids',
    'ozonesquash',
    'ozone_profile_rcemip',
    'parse_fraction_of_day',
]

logger = logging.getLogger(__name__)


def append_description(dataset, description=None):
    """Append variable attributes to a given dataset.

    Parameters:
          dataset (xarray.Dataset): Dataset including variables to describe.
          description (dict): Dictionary containing variable descriptions.
            The keys are the variable keys used in the Dataset.
            The values are dictionaries themselves containing attributes
            and their names as keys, e.g.:
                desc = {'T': {'units': 'K', 'standard_name': 'temperature'}}
    """
    if description is None:
        description = constants.variable_description

    for key in dataset.variables:
        if key in description:
            dataset[key].attrs = constants.variable_description[key]


def append_timestep_netcdf(filename, data, timestamp):
    """Append a timestep to an existing netCDF4 file.

    Notes:
        The variables to append to have to exist in the netCDF4 file!

    Parameters:
        filename (str): Path to the netCDF4.
        data (dict{ndarray}): Dict-like object containing the data arrays.
            The key is the variable name and the value is an ``ndarray``, a
            ``pandas.Series`` or an ``xarray.DataArray`` e.g.:
                >>> data = {'T': np.array([290, 295, 300])}

        timestamp (float): Timestamp of values appended.
    """
    # Open netCDF4 file in `append` mode.
    with Dataset(filename, 'a') as nc:
        logging.debug('Append timestep to "{}".'.format(filename))
        t = nc.dimensions['time'].size  # get index to store data.
        nc.variables['time'][t] = timestamp  # append timestamp.

        # Append values for each data variable in ``data``.
        for var in data.data_vars:
            # Append variable if it has a `time` dimension.
            if 'time' in nc[var].dimensions:
                # TODO: Find a cleaner way to handle different data dimensions.
                if 'plev' in nc[var].dimensions:
                    if hasattr(data[var], 'values'):
                        nc.variables[var][t, :] = data[var].values
                    else:
                        nc.variables[var][t, :] = data[var]
                else:
                    if hasattr(data[var], 'values'):
                        nc.variables[var][t] = data[var].values
                    else:
                        nc.variables[var][t] = data[var]


def return_if_type(variable, variablename, expect, default):
    """Return a variable if it matches an expected type.

    Parameters:
          variable: Variable to check.
          variablename (str): Variable name for error message.
          expect (type): Expected variable type.
          default: Default value, if varibale is ``None``.

    Raises:
          TypeError: If variable does not match expected type.
    """
    if variable is None:
        # use a surface with heat capacity as default.
        variable = default
    elif not isinstance(variable, expect):
        raise TypeError(
            'Argument `{name}` has to be of type `{type}`.'.format(
                name=variablename, type=expect.__name__)
        )

    return variable


def phlev_from_plev(fulllevels):
    """Returns the linear interpolated halflevels for given array.

    Parameters:
        fulllevels (ndarray): Pressure at fullevels.

    Returns:
        ndarray: Coordinates at halflevel.

    """
    plev_log = np.log(fulllevels)  # Perform inter-/extrapolation in log-space

    inter = 0.5 * (plev_log[1:] + plev_log[:-1])
    bottom = plev_log[0] + 0.5 * (plev_log[0] - plev_log[1])
    top = plev_log[-1] - 0.5 * (plev_log[-2] - plev_log[-1])

    return np.exp(np.hstack((bottom, inter, top)))


def dz_from_z(z):
    """
    Return the level thickness given the height array.

    Parameters:
        z (ndarray): Height of each model level [m]

    Returns:
        ndarray: Thickness of model levels [m]
    """
    dz = np.hstack([z[0], np.diff(z)])  # TODO: is this a good approx?
    return dz


def refined_pgrid(start, stop, num=200, shift=0.5, fixpoint=0.):
    """Create a pressure grid with adjustable distribution in logspace.

    Notes:
          Wrapper for ``typhon.math.squeezable_logspace``.

    Parameters:
        start (float): The starting value of the sequence.
        stop (float): The end value of the sequence.
        num (int): Number of sample to generate (Default is 50).
        shift (float): Factor with which the first stepwidth is
            squeezed in logspace. Has to be between  ``(0, 2)`.
            Values smaller than one compress the gridpoints,
            while values greater than 1 strecht the spacing.
            The default is ``0.5`` (bottom heavy.)
        fixpoint (float): Relative fixpoint for squeezing the grid.
            Has to be between ``[0, 1]``. The  default is ``0`` (bottom).

    Returns:
        ndarray: Pressure grid.
    """
    grid = ty.math.squeezable_logspace(
        start=start, stop=stop, num=num, squeeze=shift, fixpoint=fixpoint
    )

    return grid


def get_pressure_grids(start=1000e2, stop=1, num=200, squeeze=0.5):
    """Create matching pressure levels and half-levels.

    Parameters:
        start (float): Pressure of the lowest half-level (surface) [Pa].
        stop (float): Pressure of the highest half-level (TOA) [Pa].
        num (int): Number of **full** pressure levels.
        squeeze (float): Factor with which the first step width is
            squeezed in logspace. Has to be between ``(0, 2)``.
            Values smaller than one compress the half-levels,
            while values greater than 1 stretch the spacing.
            The default is ``0.5`` (bottom heavy.)

    Returns:
        ndarray, ndarray: Full-level pressure, half-level pressure [Pa].
    """
    phlev = ty.math.squeezable_logspace(start, stop, num + 1, squeeze=squeeze)
    plev = np.exp(0.5 * (np.log(phlev[1:]) + np.log(phlev[:-1])))

    return plev, phlev


def ozonesquash(o3, z, squash):
    """
    Squashes the ozone profile upwards or stretches it downwards, with no
        change to the shape of the profile above the ozone concentration maximum
    Parameters:
        o3 (ndarray): initial ozone profile
        z (ndarray): corresponding height values
        squash: float, with 1 being no squash,
            numbers < 1 squashing the profile towards the maximum,
            numbers > 1, stretching the profile downwards
    Returns:
        ndarray: new ozone profile
    """
    i_max_o3 = np.argmax(o3)

    sqz = (z[:i_max_o3] - z[i_max_o3])*squash + z[i_max_o3]
    new_o3 = copy.copy(o3)
    new_o3[:i_max_o3] = np.interp(z[:i_max_o3], sqz, o3[:i_max_o3])
    return new_o3


def ozone_profile_rcemip(plev, g1=3.6478, g2=0.83209, g3=11.3515):
    """Compute the ozone volumetric mixing ratio from pressure.

    .. math::
        O_3 = g_1 \cdot p^{g_2} e^\frac{-p}{g_3}

    Parameters:
        plev (ndarray): Atmospheric pressure [Pa].
        g1, g2, g3 (float): Fitting parameters for gamma distribution
            according to Wing et al. (2017).

    Returns:
          ndarray: Ozone profile [VMR].

    Reference:
        Wing et al., 2017, Radiative-Convective Equilibrium Model
        Intercomparison Project

    """
    p = plev / 100
    return g1 * p**g2 * np.exp(-p / g3) * 1e-6


def parse_fraction_of_day(time):
    """Calculate the fraction of a day.

    Parameters:
        time (str or float): Specified time delta (e.g. '6h').
            Valid units:
                's' for seconds
                'm' for minutes
                'h' for hours
                'd' for days
                'w' for weeks
            If numeric, return value.

    Returns:
        float: Fraction of a day.

    Example:
        >>> parse_fraction_of_day('12h')
        0.5
    """
    mapping = {
        's': 'seconds',
        'm': 'minutes',
        'h': 'hours',
        'd': 'days',
        'w': 'weeks',
    }

    if isinstance(time, str):
        value, period = float(time[:-1]), mapping[time[-1]]
        return timedelta(**{period: value}).total_seconds() / 3600 / 24
    elif isinstance(time, Number):
        return time



# -*- coding: utf-8 -*-
"""This package aims to provide a framework to run radiative-convective
equilibirum (RCE) simulations.

The package is written in a strict object - oriented way. The RCE model is
subdivided into self-contained submodels(atmosphere, surface, radiation) that
can be implemented independently from each other. This allows to exchange parts
of the model setup seperately to investigate certain effects. The different
submodels need to fulfill requirements to make interaction possible. These
requirements are enforced by the use of abstract base classes.
"""
import logging
from os.path import (join, dirname)

__version__ = open(join(dirname(__file__), 'VERSION')).read().strip()

from . import atmosphere
from . import cloud
from . import component
from . import constants
from . import convection
from . import humidity
from . import lapserate
from . import netcdf
from . import ozone
from . import plots
from . import radiation
from . import surface
from . import upwelling
from . import utils
from .core import RCE


__all__ = [
    'RCE',
    'atmosphere',
    'cloud',
    'component',
    'constants',
    'convection',
    'humidity',
    'lapserate',
    'netcdf',
    'ozone',
    'plots',
    'radiation',
    'surface',
    'upwelling',
    'utils',
]

# Basic configuration for all loggers used within konrad.
# NOTE: The process name is included for more verbose logs in multiprocessing.
logging.basicConfig(
    # filename='log.txt',
    level=logging.INFO,
    datefmt='%Y-%m-%d %H:%M:%S',
    style='{',  # Allows to use format string syntax in the next line.
    format='{asctime} {processName}:{levelname}:{name}:{message}',
    )



# -*- coding: utf-8 -*-
"""This module contains classes for handling humidity."""
import abc
import logging

import numpy as np
from typhon.atmosphere import vmr as rh2vmr
from scipy.stats import skewnorm

from konrad.component import Component


logger = logging.getLogger(__name__)


class Humidity(Component, metaclass=abc.ABCMeta):
    """Base class to define abstract methods for all humidity handlers."""
    def __init__(self, rh_surface=0.8, rh_tropo=0.3, p_tropo=170e2,
                 vmr_strato=None, vmr_profile=None, rh_profile=None):
        """Create a humidity handler.

        Parameters:
            rh_surface (float): Relative humidity at first
                pressure level (surface).
            rh_tropo (float): Relative humidity at second maximum
                in the upper-troposphere.
            p_tropo (float): Pressure level of second humidity maximum [Pa].
            vmr_strato (float): Stratospheric water vapor VMR.
            vmr_profile (ndarray): Water vapour volume mixing ratio profile.
            rh_profile (ndarray): Relative humidity profile.
                Is ignored if the `vmr_profile` is given.
        """
        self.rh_surface = rh_surface
        self.rh_tropo = rh_tropo
        self.p_tropo = p_tropo
        self.vmr_strato = vmr_strato

        self._vmr_profile = vmr_profile
        self._rh_profile = rh_profile

    @abc.abstractmethod
    def __call__(self, plev, T, z, **kwargs):
        """Determine the humidity profile based on atmospheric state.

        Parameters:
            plev (ndarray): Pressure levels [Pa].
            T (ndarray): Temperature [K].
            z (ndarray): Height [m].

        Returns:
            ndarray: Water vapor profile [VMR].
        """

    def get_relative_humidity_profile(self, p):
        """Create a realistic relative humidity profile for the tropics.

        Parameters:
            p (ndarray): Pressure [Pa].

        Returns:
            ndarray: Relative humidity.
        """
        # If set, return prescribed relative humidity profile.
        if self._rh_profile is not None:
            return self._rh_profile

        # Exponential decay from the surface value throughout the atmosphere.
        rh = self.rh_surface * (p / p[0]) ** 1.15

        # Add skew-normal distribution.
        rh += self.rh_tropo * skewnorm.pdf(
            x=np.log(p),
            # NOTE: Subtract 20e2 hPa to match the passed value of `p_tropo`
            # with the actual peak of the skewnorm distribution.
            loc=np.log(self.p_tropo - 20e2),
            # Shape parameters are fitted to an ERA5 climatology.
            a=4,
            scale=0.75,
        )

        return rh

    def get_vmr_profile(self, atmosphere):
        """Return a water vapor volume mixing ratio profile.

        Parameters:
            atmosphere (konrad.atmosphere): atmosphere object,
                including profiles of temperature [K], pressure [Pa],
                altitude [m]

        Returns:
            ndarray: Water vapor profile [VMR].
        """
        p = atmosphere.get('plev')
        T = atmosphere.get('T', keepdims=False)
        z = atmosphere.get('z', keepdims=False)

        if self._vmr_profile is None:
            vmr = rh2vmr(self.get_relative_humidity_profile(p), p, T)
        else:
            vmr = self._vmr_profile

        vmr = self.adjust_stratospheric_vmr(vmr, p, T, z)

        return vmr

    def adjust_stratospheric_vmr(self, vmr, p, T, z, cold_point_min=1e2):
        """Adjust water vapor VMR values in the stratosphere.

        The stratosphere is determined using the cold point (pressure level
        with lowest temperatures above a given pressure threshold).

        Parameters:
            vmr (ndarray): Water vapor mixing ratio [VMR].
            p (ndarray): Pressure levels [Pa].
            T (ndarray): Temperature [K].
            z (ndarray): Height [m].
            cold_point_min (float): Lower threshold for cold point pressure.

        Returns:
              ndarray: Adjusted water vapor profile [VMR].
        """
        # Determine the level index of the cold point as proxy for the
        # transition between troposphere and stratosphere. The pressure has
        # to be above a given threshold to prevent finding a cold point at
        # the top of the atmosphere.
        cp_index = int(np.argmin(T[p > cold_point_min]))

        # Keep stratospheric water vapor VMR values constant.
        # If stratospheric background = None, use cold point VMR, otherwise
        # transition to the stratospheric background value
        if self.vmr_strato is None:
            vmr[cp_index:] = vmr[cp_index]
        else:
            # If the VMR falls below the stratospheric background...
            if np.any(vmr[p > cold_point_min] < self.vmr_strato):
                # ... set all values equal to the background from there.
                vmr[np.argmax(vmr < self.vmr_strato):] = self.vmr_strato
            else:
                # Otherwise find the smallest VMR and use the background value
                # from there on, this at least minimizes the discontinuity at
                # the transition point.
                vmr[np.argmin(vmr[p > cold_point_min]):] = self.vmr_strato

        return vmr


class FixedVMR(Humidity):
    """Keep the water vapor volume mixing ratio constant."""
    def __call__(self, atmosphere, **kwargs):
        if self._vmr_profile is None:
            self._vmr_profile = self.get_vmr_profile(atmosphere)

        return self.get_vmr_profile(atmosphere)


class FixedRH(Humidity):
    """Preserve the relative humidity profile under temperature changes.

    The relative humidity is kept constant under temperature changes,
    allowing for a moistening in a warming climate.
    """
    def __call__(self, atmosphere, **kwargs):
        return self.get_vmr_profile(atmosphere)


class Manabe67(Humidity):
    """Relative humidity model following Manabe and Wetherald (1967)."""
    def __init__(self, rh_surface=0.77, vmr_strato=4.8e-6):
        """Initialize a humidity model.

        Parameters:
            rh_surface (float): Relative humidity at the surface.
            vmr_strato (float): Minimum water vapor volume mixing ratio.
                Values below this threshold are clipped. The default value
                of `4.8e-6` resembles the *mass* mixing ratio of 3 ppm given
                in the literature.
        """
        super().__init__(rh_surface=rh_surface, vmr_strato=vmr_strato)

    def get_relative_humidity_profile(self, p):
        return self.rh_surface * (p / p[0] - 0.02) / (1 - 0.02)

    def __call__(self, atmosphere, **kwargs):
        return self.get_vmr_profile(atmosphere)


class Cess76(FixedRH):
    """Relative humidity model following Cess (1976).

    The relative humidity profile depends on the surface temperature.
    This results in moister atmospheres at warmer temperatures.
    """
    def __init__(self, rh_surface=0.8, T_surface=288):
        """Initialize a humidity model.

        Parameters:
            rh_surface (float): Relative humidity at the surface.
            T_surface (float): Surface temperature [K].
        """
        super().__init__(rh_surface=rh_surface, rh_tropo=0.)

        self.T_surface = T_surface

    @property
    def omega(self):
        """Temperature dependent scaling factor for the RH profile."""
        return 1.0 - 0.03 * (self.T_surface - 288)

    def get_relative_humidity_profile(self, p):
        return self.rh_surface * ((p / p[0] - 0.02) / (1 - 0.02))**self.omega

    def __call__(self, atmosphere, surface, **kwargs):
        """Determine the humidity profile based on atmospheric state.

        Parameters:
            atmosphere (konrad.atmosphere): atmosphere object,
                including temperature, pressure, altitude
            surface (konrad.surface): Surface object

        Returns:
            ndarray: Water vapor profile [VMR].
        """
        self.T_surface = surface['temperature'][-1]

        return self.get_vmr_profile(atmosphere)


class CoupledRH(Humidity):
    """Couple the relative humidity profile to the top of convection.

    The relative humidity is kept constant under temperature changes,
    allowing for a moistening in a warming climate. In addition,
    the vertical structure of the humidity profile is coupled to the
    structure of atmospheric convection (Zelinka et al, 2010).

    References:
        Zelinka and Hartmann, 2010, Why is longwave cloud feedback positive?
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # This ensures a proper coupling of the relative humidity profile.
        for attr in ('_rh_profile', '_vmr_profile'):
            if getattr(self, attr) is not None:
                logger.warning(
                    'Set attribute "{}" to `None` for coupled '
                    'humidity calculations.'.format(attr)
                )
                setattr(self, attr, None)

    def __call__(self, atmosphere, convection, **kwargs):
        """Determine the humidity profile based on atmospheric state.

        Parameters:
            atmosphere (konrad.atmosphere): atmosphere object,
                including temperature, pressure, altitude
            convection (konrad.convection): convection scheme

        Returns:
            ndarray: Water vapor profile [VMR].
        """
        p_tropo = convection.get('convective_top_plev')[0]
        if p_tropo is not None and not np.isnan(p_tropo):
            self.p_tropo = p_tropo

        return self.get_vmr_profile(atmosphere)



# -*- coding: utf-8 -*-
"""This module contains classes for handling clouds."""
import abc
import logging

import numpy as np
from scipy.interpolate import interp1d
from sympl import DataArray
from konrad.utils import dz_from_z

logger = logging.getLogger(__name__)


def get_p_data_array(values, units='kg m^-2', numlevels=200):
    """Return a DataArray of values."""
    if type(values) is DataArray:
        return values

    elif type(values) is int or type(values) is float:
        return DataArray(values*np.ones(numlevels,),
                         dims=('mid_levels',),
                         attrs={'units': units})

    raise TypeError(
            'Cloud variable input must be a single value or a sympl.DataArray')


def get_waveband_data_array(values, units='dimensionless', numbands=14,
                            numlevels=200, sw=True):
    """Return a DataArray of values."""
    if type(values) is DataArray:
        return values

    dims_bands = 'num_shortwave_bands' if sw else 'num_longwave_bands'

    if type(values) is int or type(values) is float:
        return DataArray(values*np.ones((numlevels, numbands)),
                         dims=('mid_levels', dims_bands),
                         attrs={'units': units})

    raise TypeError(
            'Cloud variable input must be a single value or a sympl.DataArray')


class Cloud(metaclass=abc.ABCMeta):
    """Base class to define abstract methods for all cloud handlers.
    Default properties include a cloud area fraction equal to zero everywhere
    (ie no cloud)."""
    num_longwave_bands = 16
    num_shortwave_bands = 14

    def __init__(self, z,
                 mass_water=0, mass_ice=0, cloud_fraction=0,
                 ice_particle_size=20, droplet_radius=10,
                 lw_optical_thickness=0, sw_optical_thickness=0,
                 forward_scattering_fraction=0, asymmetry_parameter=0.85,
                 single_scattering_albedo=0.9):
        """Create a cloud.

        Parameters:
            z (ndarray): array of height values [m]
            mass_water (float / DataArray): mass content of cloud liquid water
                [kg m-2]
            mass_ice (float / DataArray): mass content of cloud ice [kg m-2]
            cloud_fraction (float / DataArray): cloud area fraction
            ice_particle_size (float / DataArray): cloud ice particle size
                [micrometers]
            droplet_radius (float / DataArray): cloud water droplet radius
                [micrometers]
            lw_optical_thickness (float / DataArray): longwave optical
                thickness of the cloud
            sw_optical_thickness (float / DataArray): shortwave optical
                thickness of the cloud
            forward_scattering_fraction (float / DataArray): cloud forward
                scattering fraction (for the shortwave component of RRTMG)
                This is a scaling factor for the other shortwave parameters,
                if it is set to 0, no scaling is applied.
            asymmetry_parameter (float / DataArray): cloud asymmetry parameter
                (for the shortwave component of RRTMG)
            single_scattering_albedo (float / DataArray): single scattering
                albedo due to cloud (for the shortwave component of RRTMG)
        """
        numlevels = z.size

        self.mass_content_of_cloud_liquid_water_in_atmosphere_layer = \
            get_p_data_array(mass_water, numlevels=numlevels)

        self.mass_content_of_cloud_ice_in_atmosphere_layer = get_p_data_array(
                mass_ice, numlevels=numlevels)

        self.cloud_area_fraction_in_atmosphere_layer = get_p_data_array(
                cloud_fraction, numlevels=numlevels, units='dimensionless')

        self.cloud_ice_particle_size = get_p_data_array(ice_particle_size,
                                                        numlevels=numlevels,
                                                        units='micrometers')

        self.cloud_water_droplet_radius = get_p_data_array(droplet_radius,
                                                           numlevels=numlevels,
                                                           units='micrometers')

        self.longwave_optical_thickness_due_to_cloud = get_waveband_data_array(
                lw_optical_thickness, numlevels=numlevels,
                numbands=self.num_longwave_bands, sw=False)

        self.cloud_forward_scattering_fraction = get_waveband_data_array(
                forward_scattering_fraction, numlevels=numlevels,
                numbands=self.num_shortwave_bands)

        self.cloud_asymmetry_parameter = get_waveband_data_array(
                asymmetry_parameter, numlevels=numlevels,
                numbands=self.num_shortwave_bands)

        self.shortwave_optical_thickness_due_to_cloud = \
            get_waveband_data_array(sw_optical_thickness,
                                    numlevels=numlevels,
                                    numbands=self.num_shortwave_bands)

        self.single_scattering_albedo_due_to_cloud = get_waveband_data_array(
                single_scattering_albedo, numlevels=numlevels,
                numbands=self.num_shortwave_bands)

    def interpolation_function(self, cloud_parameter, z):
        """ Calculate the interpolation function, to be used to keep the
        cloud at a constant thickness [m] and attached to a normalisation
        level (self._norm_level). A separate interpolation function is required
        for each cloud parameter that needs to be interpolated.

        Parameters:
            cloud_parameter (DataArray): property to be interpolation
            z (ndarray): height array, on which the interpolation is performed
        Returns:
            scipy.interpolate.interpolate.interp1d
        """
        normed_height = z - self._norm_level
        interpolation_f = interp1d(
            normed_height,
            cloud_parameter.values,
            axis=0,
            fill_value='extrapolate',
        )
        return interpolation_f

    def shift_property(self, cloud_parameter, interpolation_f, z, norm_new):
        """Shift the cloud area fraction according to a normalisation level.
        Maintain the thickness of the cloud [m].

        Parameters:
            cloud_parameter (DataArray): cloud property to be shifted
            interpolation_f (scipy.interpolate.interpolate.interp1d):
                interpolation object calculated by interpolation_function
            z (ndarray): height array [m]
            norm_new (int / float): normalisation height [m]

        Returns:
            DataArray: shifted cloud property
        """
        if norm_new is not np.nan:
            # Move the cloud to the new normalisation level, if there is one.
            cloud_parameter.values = interpolation_f(z - norm_new)
        else:
            # Otherwise keep the cloud where it is.
            cloud_parameter.values = interpolation_f(z - self._norm_level)

        return cloud_parameter

    def scale_optical_thickness(self, dz):
        """Conserve the total optical depth of the cloud.
        If we assume that the cloud remains at the same altitude and with the
        same thickness, then we also want to assume that the optical depth of
        the cloud remains the same.
        This is equivalent to conservation of cloud mass, as optical depth is
        proportional to density.

        Parameters:
            dz (ndarray): thickness of model levels [m]
        """
        dz_sw_array = np.repeat(dz[:, np.newaxis], self.num_shortwave_bands,
                                axis=1)
        dz_lw_array = np.repeat(dz[:, np.newaxis], self.num_longwave_bands,
                                axis=1)

        self.longwave_optical_thickness_due_to_cloud = DataArray(
            dz_lw_array * self._longwave_optical_thickness_per_meter,
            dims=('mid_levels', 'num_longwave_bands'),
            attrs={'units': 'dimensionless'}
        )
        self.shortwave_optical_thickness_due_to_cloud = DataArray(
            dz_sw_array * self._shortwave_optical_thickness_per_meter,
            dims=('mid_levels', 'num_shortwave_bands'),
            attrs={'units': 'dimensionless'}
        )

    @staticmethod
    def calculate_mass_cloud(cloud_fraction_array, dz, density):
        """Calculate the mass of the cloud on each model level given the cloud
        area fraction and the density within the cloud.

        Parameters:
            cloud_fraction_array (ndarray): cloud area fraction
            dz (ndarray): thickness of model levels [m]
            density (int / float): density of cloud mass [g m^-3]

        Returns:
            DataArray: mass of cloud on each model level [kg m^-2]
        """
        mass_array = cloud_fraction_array * density * 10 ** -3 * dz
        mass = DataArray(mass_array, dims=('mid_levels',),
                         attrs={'units': 'kg m^-2'})
        return mass

    @abc.abstractmethod
    def update_cloud_profile(self, atmosphere, convection, **kwargs):
        """Return the cloud parameters for the radiation scheme.
        Parameters:
            atmosphere (konrad.atmosphere.Atmosphere): atmosphere model
            convection (konrad.convection): convection scheme
        """


class ClearSky(Cloud):
    def update_cloud_profile(self, *args, **kwargs):
        return


class DirectInputCloud(Cloud):
    """ To be used with cloud_optical_properties='direct_input' in climt/RRTMG.
    """
    def __init__(self, z, cloud_fraction,
                 lw_optical_thickness, sw_optical_thickness,
                 forward_scattering_fraction=0, asymmetry_parameter=0.85,
                 single_scattering_albedo=0.9):

        super().__init__(
            z,
            cloud_fraction=cloud_fraction,
            lw_optical_thickness=lw_optical_thickness,
            sw_optical_thickness=sw_optical_thickness,
            forward_scattering_fraction=forward_scattering_fraction,
            asymmetry_parameter=asymmetry_parameter,
            single_scattering_albedo=single_scattering_albedo
        )

        dz = dz_from_z(z)
        dz_sw_array = np.repeat(dz[:, np.newaxis], self.num_shortwave_bands,
                                axis=1)
        dz_lw_array = np.repeat(dz[:, np.newaxis], self.num_longwave_bands,
                                axis=1)
        self._longwave_optical_thickness_per_meter = \
            self.longwave_optical_thickness_due_to_cloud / dz_lw_array
        self._shortwave_optical_thickness_per_meter = \
            self.shortwave_optical_thickness_due_to_cloud / dz_sw_array

        self._norm_level = 0
        self._f = self.interpolation_function(
            self.cloud_area_fraction_in_atmosphere_layer, z)
        self._interpolation_sw = self.interpolation_function(
            self._shortwave_optical_thickness_per_meter, z)
        self._interpolation_lw = self.interpolation_function(
            self._longwave_optical_thickness_per_meter, z)

    def update_cloud_profile(self, atmosphere, **kwargs):
        """ Keep the cloud profile fixed with height.
        """
        z = atmosphere.get('z', keepdims=False)
        dz = dz_from_z(z)

        self.cloud_area_fraction_in_atmosphere_layer = self.shift_property(
            self.cloud_area_fraction_in_atmosphere_layer, self._f, z, 0)

        self._shortwave_optical_thickness_per_meter = self.shift_property(
            self._shortwave_optical_thickness_per_meter,
            self._interpolation_sw, z, 0)
        self._longwave_optical_thickness_per_meter = self.shift_property(
            self._longwave_optical_thickness_per_meter,
            self._interpolation_lw, z, 0)

        self.scale_optical_thickness(dz)


class HighCloud(Cloud):
    """ To be used with cloud_optical_properties='liquid_and_ice_clouds' in
    climt/RRTMG."""
    def __init__(self, z, cloud_top=12000, depth=500,
                 cloud_fraction_in_cloud=1, ice_density=0.5,
                 ice_particle_size=20):
        """
        Parameters:
            z (ndarray): altitude values [m]
            cloud_top (int): Cloud top height [m]
                This should be the convective top
            depth (int): Cloud depth / thickness of cloud [m]
            cloud_fraction_in_cloud (int / float): Cloud area fraction of the
                cloud for the model levels between cloud_top - depth and
                cloud_top
            ice_density (int / float): density of cloud ice [g m-3]
            ice_particle_size (int / float): Cloud ice particle size [microns]
        """
        cloud_fraction_array = np.zeros(z.shape)
        cloud_fraction_array[(z < cloud_top) & (
                z > cloud_top-depth)] = cloud_fraction_in_cloud
        cloud_fraction = DataArray(cloud_fraction_array, dims=('mid_levels',),
                                   attrs={'units': 'dimensionless'})

        mass_ice = self.calculate_mass_cloud(cloud_fraction_array, dz_from_z(z),
                                             ice_density)

        super().__init__(
            z,
            cloud_fraction=cloud_fraction,
            mass_ice=mass_ice,
            ice_particle_size=ice_particle_size
        )

        self._ice_density = ice_density
        self._norm_level = cloud_top
        self._f = self.interpolation_function(
            self.cloud_area_fraction_in_atmosphere_layer, z)

    def update_cloud_profile(self, atmosphere, convection, **kwargs):
        """ Keep the cloud attached to the convective top.
        """
        norm_new = convection.get('convective_top_height')[0]
        z = atmosphere.get('z', keepdims=False)
        dz = dz_from_z(z)

        self.cloud_area_fraction_in_atmosphere_layer = self.shift_property(
            self.cloud_area_fraction_in_atmosphere_layer, self._f, z, norm_new)

        mass_ice = self.calculate_mass_cloud(
            self.cloud_area_fraction_in_atmosphere_layer.values,
            dz,
            self._ice_density)
        self.mass_content_of_cloud_ice_in_atmosphere_layer = mass_ice


class LowCloud(Cloud):
    """ To be used with cloud_optical_properties='liquid_and_ice_clouds' in
    climt/RRTMG."""
    def __init__(self, z, cloud_top=2000, depth=1500,
                 cloud_fraction_in_cloud=1, water_density=0.5,
                 droplet_radius=10):
        """
        Parameters:
            z (ndarray): altitude values [m]
            cloud_top (int): Cloud top height [m]
            depth (int): Cloud depth / thickness of cloud [m]
            cloud_fraction_in_cloud (int): Cloud area fraction of the cloud for
                the model levels between cloud_top-depth and cloud_top
            water_density (int / float): density of cloud water [g m-3]
            droplet_radius (int / float): Cloud water droplet radius [microns]
        """
        cloud_fraction_array = np.zeros(z.shape)
        cloud_fraction_array[(z < cloud_top) & (
                z > cloud_top-depth)] = cloud_fraction_in_cloud
        cloud_fraction = DataArray(cloud_fraction_array, dims=('mid_levels',),
                                   attrs={'units': 'dimensionless'})

        mass_water = self.calculate_mass_cloud(cloud_fraction_array,
                                               dz_from_z(z), water_density)

        super().__init__(
            z,
            cloud_fraction=cloud_fraction,
            mass_water=mass_water,
            droplet_radius=droplet_radius
        )

        self._water_density = water_density
        self._f = self.interpolation_function(
            self.cloud_area_fraction_in_atmosphere_layer, z)
        self._norm_level = 0

    def update_cloud_profile(self, atmosphere, **kwargs):
        """ Keep the cloud fixed with height, in the planetary boundary layer.
        """
        z = atmosphere.get('z', keepdims=False)
        dz = dz_from_z(z)

        self.cloud_area_fraction_in_atmosphere_layer = self.shift_property(
            self.cloud_area_fraction_in_atmosphere_layer, self._f, z, 0)

        mass = self.calculate_mass_cloud(
            self.cloud_area_fraction_in_atmosphere_layer.values,
            dz,
            self._water_density)
        self.mass_content_of_cloud_liquid_water_in_atmosphere_layer = mass



"""Define an interface for the RRTMG radiation scheme (through CliMT). """
import numpy as np
import datetime
from sympl import DataArray
from typhon.physics import vmr2specific_humidity
import climt

from .radiation import Radiation
from konrad.cloud import ClearSky

__all__ = [
    'RRTMG',
]


class RRTMG(Radiation):
    """RRTMG radiation scheme using the CliMT python wrapper."""

    def __init__(self, *args, solar_constant=510, mcica=False,
                 cloud_ice_properties='ebert_curry_two', **kwargs):
        super().__init__(*args, **kwargs)
        self._state_lw = None
        self._state_sw = None

        self._rad_lw = None
        self._rad_sw = None

        self._mcica = mcica
        self._cloud_ice_properties = cloud_ice_properties

        self.solar_constant = solar_constant

    def init_radiative_state(self, atmosphere):

        climt.set_constants_from_dict({"stellar_irradiance": {
                "value": self.solar_constant, "units": 'W m^-2'}})
        self._rad_lw = climt.RRTMGLongwave(
            cloud_optical_properties='liquid_and_ice_clouds',
            cloud_ice_properties=self._cloud_ice_properties,
            cloud_liquid_water_properties='radius_dependent_absorption',
            mcica=self._mcica)
        self._rad_sw = climt.RRTMGShortwave(
            ignore_day_of_year=True,
            cloud_optical_properties='liquid_and_ice_clouds',
            cloud_ice_properties=self._cloud_ice_properties,
            cloud_liquid_water_properties='radius_dependent_absorption',
            mcica=self._mcica)
        state_lw = {}
        state_sw = {}

        plev = atmosphere['plev']
        phlev = atmosphere['phlev']
        numlevels = len(plev)

        for state0 in state_lw, state_sw:
            state0['mid_levels'] = DataArray(
                    np.arange(0, numlevels),
                    dims=('mid_levels',),
                    attrs={'label': 'mid_levels'})

            state0['interface_levels'] = DataArray(
                    np.arange(0, numlevels+1),
                    dims=('interface_levels',),
                    attrs={'label': 'interface_levels'})

            state0['air_pressure'] = DataArray(
                    plev,
                    dims=('mid_levels',),
                    attrs={'units': 'Pa'})

            state0['air_pressure_on_interface_levels'] = DataArray(
                    phlev,
                    dims=('interface_levels',),
                    attrs={'units': 'Pa'})

            state0['time'] = datetime.datetime(2000, 1, 1)

        ### Aerosols ###
        # TODO: Should all the aerosol values be zero?!
        # Longwave specific
        num_lw_bands = self._rad_lw.num_longwave_bands

        state_lw['longwave_optical_thickness_due_to_aerosol'] = DataArray(
                np.zeros((num_lw_bands, numlevels)),
                dims=('num_longwave_bands', 'mid_levels'),
                attrs={'units': 'dimensionless'})
        state_lw['surface_longwave_emissivity'] = DataArray(
            np.ones((num_lw_bands,)),
            dims=('num_longwave_bands'),
            attrs={'units': 'dimensionless'})

        # Shortwave specific changes
        num_sw_bands = self._rad_sw.num_shortwave_bands

        num_aerosols = self._rad_sw.num_ecmwf_aerosols
        state_sw['aerosol_optical_depth_at_55_micron'] = DataArray(
                np.zeros((num_aerosols, numlevels)),
                dims=('num_ecmwf_aerosols', 'mid_levels'),
                attrs={'units': 'dimensionless'})
        state_sw['solar_cycle_fraction'] = DataArray(
            0, attrs={'units': 'dimensionless'})
        state_sw['flux_adjustment_for_earth_sun_distance'] = DataArray(
            1, attrs={'units': 'dimensionless'}
        )

        for quant in ['shortwave_optical_thickness_due_to_aerosol',
                      'single_scattering_albedo_due_to_aerosol',
                      'aerosol_asymmetry_parameter']:
            state_sw[quant] = DataArray(
                np.zeros((num_sw_bands, numlevels)),
                dims=('num_shortwave_bands', 'mid_levels'),
                attrs={'units': 'dimensionless'})

        return state_lw, state_sw

    def update_cloudy_radiative_state(self, cloud, state0, sw=True):

        # Take cloud quantities from cloud class.
        state0['cloud_ice_particle_size'] = cloud.cloud_ice_particle_size
        state0['mass_content_of_cloud_liquid_water_in_atmosphere_layer'] = cloud.mass_content_of_cloud_liquid_water_in_atmosphere_layer
        state0['mass_content_of_cloud_ice_in_atmosphere_layer'] = cloud.mass_content_of_cloud_ice_in_atmosphere_layer
        state0['cloud_water_droplet_radius'] = cloud.cloud_water_droplet_radius
        state0['cloud_area_fraction_in_atmosphere_layer'] = cloud.cloud_area_fraction_in_atmosphere_layer

        if not sw:
            state0['longwave_optical_thickness_due_to_cloud'] = cloud.longwave_optical_thickness_due_to_cloud
        else:
            state0['cloud_forward_scattering_fraction'] = cloud.cloud_forward_scattering_fraction
            state0['cloud_asymmetry_parameter'] = cloud.cloud_asymmetry_parameter
            state0['shortwave_optical_thickness_due_to_cloud'] = cloud.shortwave_optical_thickness_due_to_cloud
            state0['single_scattering_albedo_due_to_cloud'] = cloud.single_scattering_albedo_due_to_cloud

        return

    def update_radiative_state(self, atmosphere, surface, state0, sw=True):
        """ Update CliMT formatted atmospheric state using parameters from our
        model.

        Parameters:
            atmosphere (konrad.atmosphere.Atmosphere): Atmosphere model.
            surface (konrad.surface): Surface model.
            state0 (dictionary): atmospheric state in the format for climt
            sw (bool): Toggle between shortwave and longwave calculations.

        Returns:
            dictionary: updated state
        """

        state0['air_temperature'] = DataArray(
            atmosphere['T'][0, :],
            dims=('mid_levels',),
            attrs={'units': 'degK'})

        vmr_h2o = atmosphere['H2O'][0, :]
        specific_humidity = vmr2specific_humidity(vmr_h2o)
        state0['specific_humidity'] = DataArray(
            specific_humidity,
            dims=('mid_levels',),
            attrs={'units': 'g/g'})

        # CliMT/konrad name mapping
        gas_name_mapping = [
            ('mole_fraction_of_methane_in_air', 'CH4'),
            ('mole_fraction_of_carbon_dioxide_in_air', 'CO2'),
            ('mole_fraction_of_nitrous_oxide_in_air', 'N2O'),
            ('mole_fraction_of_ozone_in_air', 'O3'),
            ('mole_fraction_of_cfc11_in_air', 'CFC11'),
            ('mole_fraction_of_cfc12_in_air', 'CFC12'),
            ('mole_fraction_of_cfc22_in_air', 'CFC22'),
            ('mole_fraction_of_carbon_tetrachloride_in_air', 'CCl4'),
            ('mole_fraction_of_oxygen_in_air', 'O2'),
        ]

        for climt_key, konrad_key in gas_name_mapping:
            state0[climt_key] = DataArray(
                atmosphere.get(konrad_key, default=0, keepdims=False),
                dims=('mid_levels',),
                attrs={'units': 'mole/mole'})

        # Surface quantities
        state0['surface_temperature'] = DataArray(
            np.array(surface['temperature'][-1]),
            attrs={'units': 'degK'})

        if sw:  # surface properties required only for shortwave
            for surface_albedo in ['surface_albedo_for_diffuse_near_infrared',
                                   'surface_albedo_for_direct_near_infrared',
                                   'surface_albedo_for_diffuse_shortwave',
                                   'surface_albedo_for_direct_shortwave']:
                state0[surface_albedo] = DataArray(
                    np.array(float(surface.albedo)),
                    attrs={'units': 'dimensionless'})

            # Sun
            state0['zenith_angle'] = DataArray(
                np.array(np.deg2rad(self.current_solar_angle)),
                attrs={'units': 'radians'})

        return state0

    def radiative_fluxes(self, atmosphere, surface, cloud):
        """Returns shortwave and longwave fluxes and heating rates.

        Parameters:
            atmosphere (konrad.atmosphere.Atmosphere): atmosphere model
            surface (konrad.surface): surface model
            cloud (konrad.cloud): cloud model

        Returns:
            tuple: containing two dictionaries, one of air temperature
            values and the other of fluxes and heating rates
        """
        if self._state_lw is None or self._state_sw is None:  # first time only
            self._state_lw, self._state_sw = self.init_radiative_state(
                    atmosphere)
            self.update_cloudy_radiative_state(cloud, self._state_lw, sw=False)
            self.update_cloudy_radiative_state(cloud, self._state_sw, sw=True)

        # if there are clouds update the cloud properties for the radiation
        if not isinstance(cloud, ClearSky):
            self.update_cloudy_radiative_state(cloud, self._state_lw, sw=False)
            self.update_cloudy_radiative_state(cloud, self._state_sw, sw=True)

        self.update_radiative_state(atmosphere, surface, self._state_lw,
                                    sw=False)
        self.update_radiative_state(atmosphere, surface, self._state_sw,
                                    sw=True)

        lw_fluxes = self._rad_lw(self._state_lw)
        sw_fluxes = self._rad_sw(self._state_sw)

        return lw_fluxes, sw_fluxes

    def calc_radiation(self, atmosphere, surface, cloud):
        """Updates the shortwave, longwave and net heatingrates.
        Converts output from radiative_fluxes to be in the format required for
        our model.

        Parameters:
            atmosphere (konrad.atmosphere.Atmosphere): Atmosphere model.
            surface (konrad.surface): Surface model.
            cloud (konrad.cloud): cloud model
        """
        lw_dT_fluxes, sw_dT_fluxes = self.radiative_fluxes(atmosphere, surface,
                                                           cloud)
        lw_fluxes = lw_dT_fluxes[1]
        sw_fluxes = sw_dT_fluxes[1]

        self['lw_htngrt'] = np.expand_dims(
                lw_fluxes['air_temperature_tendency_from_longwave'].data, 0)
        self['lw_htngrt_clr'] = np.expand_dims(
                lw_fluxes['air_temperature_tendency_from_longwave_assuming_clear_sky'].data, 0)
        self['lw_flxu'] = np.expand_dims(
                lw_fluxes['upwelling_longwave_flux_in_air'].data, 0)
        self['lw_flxd'] = np.expand_dims(
                lw_fluxes['downwelling_longwave_flux_in_air'].data, 0)
        self['lw_flxu_clr'] = np.expand_dims(
                lw_fluxes['upwelling_longwave_flux_in_air_assuming_clear_sky'].data, 0)
        self['lw_flxd_clr'] = np.expand_dims(
                lw_fluxes['downwelling_longwave_flux_in_air_assuming_clear_sky'].data, 0)
        self['sw_htngrt'] = np.expand_dims(
                sw_fluxes['air_temperature_tendency_from_shortwave'].data, 0)
        self['sw_htngrt_clr'] = np.expand_dims(
                sw_fluxes['air_temperature_tendency_from_shortwave_assuming_clear_sky'].data, 0)
        self['sw_flxu'] = np.expand_dims(
                sw_fluxes['upwelling_shortwave_flux_in_air'].data, 0)
        self['sw_flxd'] = np.expand_dims(
                sw_fluxes['downwelling_shortwave_flux_in_air'].data, 0)
        self['sw_flxu_clr'] = np.expand_dims(
                sw_fluxes['upwelling_shortwave_flux_in_air_assuming_clear_sky'].data, 0)
        self['sw_flxd_clr'] = np.expand_dims(
                sw_fluxes['downwelling_shortwave_flux_in_air_assuming_clear_sky'].data, 0)

        self.coords={
            'time': np.array([0]),
            'phlev': atmosphere['phlev'],
            'plev': atmosphere['plev'],
        }



# -*- coding: utf-8 -*-
"""Plotting related functions.
"""
from matplotlib.ticker import FuncFormatter
import matplotlib.pyplot as plt
import numpy as np
import typhon.plots


__all__ = [
    'plot_overview_p_log',
    'plot_overview_z',
    'gregory_plot',
]


@FuncFormatter
def _percent_formatter(x, pos):
    return '{:.0f}\N{SIX-PER-EM SPACE}%'.format(x * 100)


def plot_overview_p_log(data, lw_htngrt, sw_htngrt, axes, **kwargs):
    """Plot overview of atmopsheric temperature and humidity profiles.

    Parameters:
        data:
        lw_htngrt:
        sw_htngrt:
        axes (list, tuple or ndarray): Three AxesSubplots.
        **kwargs: Additional keyword arguments passed to all calls
            of `atmospheric_profile`.
    """
    if len(axes) != 3:
        raise Exception('Need to pass three AxesSubplot.')
    ax1, ax2, ax3 = np.ravel(axes)

    # Plot temperature, ...
    typhon.plots.profile_p_log(data['plev'], data['T'].ravel(),
                               ax=ax1, **kwargs)
    ax1.set_xlabel('Temperature [K]')
    ax1.set_xlim(140, 320)

    # ... water vapor ...
    typhon.plots.profile_p_log(data['plev'], data['H2O'].ravel(),
                               ax=ax2, **kwargs)
    ax2.set_xlabel('$\mathsf{H_2O}$ [VMR]')
    ax2.set_xlim(0, 0.04)

    typhon.plots.profile_p_log(data['plev'], lw_htngrt,
                               ax=ax3, label='Longwave')
    typhon.plots.profile_p_log(data['plev'], sw_htngrt,
                               ax=ax3, label='Shortwave')
    typhon.plots.profile_p_log(data['plev'], sw_htngrt + lw_htngrt, ax=ax3,
                               label='Net rate', color='k')
    ax3.set_xlabel('Heatingrate [°C/day]')
    ax3.legend(loc='upper center')


def plot_overview_z(data, lw_htngrt, sw_htngrt, axes, **kwargs):
    """Plot overview of atmopsheric temperature and humidity profiles.

    Parameters:
        data:
        lw_htngrt:
        sw_htngrt:
        axes (list, tuple or ndarray): Three AxesSubplots.
        **kwargs: Additional keyword arguments passed to all calls
            of `atmospheric_profile`.
    """
    if len(axes) != 3:
        raise Exception('Need to pass three AxesSubplot.')
    ax1, ax2, ax3 = np.ravel(axes)

    # Plot temperature, ...
    typhon.plots.profile_z(data['Z'], data['T'], ax=ax1, **kwargs)
    ax1.set_xlabel('Temperature [K]')
    ax1.set_xlim(140, 320)

    # ... water vapor ...
    typhon.plots.profile_z(data['Z'], data['H2O'], ax=ax2, **kwargs)
    ax2.set_xlabel('$\mathsf{H_2O}$ [VMR]')
    ax2.set_xlim(0, 0.04)

    typhon.plots.profile_z(data['Z'], lw_htngrt, ax=ax3, label='Longwave')
    typhon.plots.profile_z(data['Z'], sw_htngrt, ax=ax3, label='Shortwave')
    typhon.plots.profile_z(data['Z'], sw_htngrt + lw_htngrt, ax=ax3,
                           label='Net rate', color='k')
    ax3.set_xlabel('Heatingrate [°C/day]')
    ax3.legend(loc='upper center')


def gregory_plot(temperature, forcing, ax=None, **kwargs):
    """Gregory plot to estimate climate sensitivity.

    Parameters:
          temperature (ndarray): Surface temperatures [K].
          forcing (ndarray): Radiation budget at top of atmosphere [W//m^2].
          ax (AxesSubplot): Axes to plot in.
          **kwargs: Additional keyword arguments are passed to ``plt.plot``.

    Returns:
          float, float: Estimated climate sensitivty [K / (W/m^2)],
            effective forcing [W / m^2].

    Examples:
        >>> temperature = np.linspace(300, 304, 25)
        >>> forcing = np.linspace(3.7, 0., temperature.size)
        >>> gregory_plot(temperature, forcing)
        (1.0810810810810814, 3.6999999999999993)

    """
    # If no axis is passed, use current axis (will create one, if needed).
    if ax is None:
        ax = plt.gca()

    # Default keyword arguments to control the plot appearance later.
    default_kwargs = {
        'marker': 'o',
        'linestyle': 'none',
    }
    default_kwargs.update(kwargs)

    # Convert temperatures into change in temperature.
    t_change = temperature - temperature[0]

    # Plot radiative forcing against surface temperature change.
    line, = ax.plot(t_change, forcing, **default_kwargs)
    ax.set_xlabel('Surface temperature change [K]')
    ax.set_ylabel('Radiation budget TOA [W/m$^2$]')
    ax.grid(True)

    # Find the maximum radiative forcing. This way, possible adjustment
    # processes during the first timesteps are not taken into account.
    max_forcing = np.argmax(forcing)
    sensitivity, eff_forcing = np.polyfit(t_change[max_forcing:],
                                          forcing[max_forcing:], 1)

    # Create x-values for "Gregory" fit. Values are plotted between zero
    # temperature change and zero forcing.
    x = np.linspace(0, - eff_forcing / sensitivity, 10)

    # Plot linear "Gregory" fit into same axis.
    ax.plot(x, eff_forcing + sensitivity * x, color=line.get_color())

    # Return the estimated climate sensivity in units of K / (W/m^2) and the
    # effective forcing in W / m^2.
    return -1 / sensitivity, eff_forcing



# -*- coding: utf-8 -*-
"""Module containing classes describing different surface models.
"""
import abc
import logging

import netCDF4
import numpy as np

from . import constants
from konrad.component import Component


__all__ = [
    'Surface',
    'SurfaceFixedTemperature',
    'SurfaceHeatCapacity',
    'SurfaceHeatSink',
]


logger = logging.getLogger(__name__)


class Surface(Component, metaclass=abc.ABCMeta):
    """Abstract base class to define requirements for surface models."""
    def __init__(self, albedo=0.2, temperature=288., height=0.):
        """Initialize a surface model.

        Parameters:
            albedo (float): Surface albedo. The default value of 0.2 is a
                decent choice for clear-sky simulation in the tropics.
            temperature (float): Surface temperature [K].
            height (float): Surface height [m].
        """
        self.albedo = albedo
        self.height = height
        self['temperature'] = (('time',), np.array([temperature]))

        # The surface pressure is initialized before the first iteration
        # within the RCE framework to ensure a pressure that is consistent
        # with the atmosphere used.
        self.pressure = None

        self.coords = {
            'time': np.array([]),
        }

    @abc.abstractmethod
    def adjust(self, sw_down, sw_up, lw_down, lw_up, timestep):
        """Adjust the surface according to given radiative fluxes.

        Parameters:
            sw_down (float): Shortwave downward flux [W / m**2].
            sw_up (float): Shortwave upward flux [W / m**2].
            lw_down (float): Longwave downward flux [W / m**2].
            lw_up (float): Longwave upward flux [W / m**2].
            timestep (float): Timestep in days.
        """
        pass

    @classmethod
    def from_atmosphere(cls, atmosphere, **kwargs):
        """Initialize a Surface object using the lowest atmosphere layer.

        Parameters:
            atmosphere (konrad.atmosphere.Atmosphere): Atmosphere model.
        """
        # Extrapolate surface height from geopotential height of lowest two
        # atmospheric layers.
        atmosphere.calculate_height()
        z = atmosphere['z'][0, :]
        z_sfc = z[0] + 0.5 * (z[0] - z[1])

        # Calculate the surface temperature following a linear lapse rate.
        # This prevents "jumps" after the first iteration, when the
        # convective adjustment is applied.
        # TODO: Perform linear or quadratic interpolation of T profile.
        lapse = 0.0065
        t_sfc = atmosphere['T'][0, 0] + lapse * (z[0] - z_sfc)

        return cls(temperature=t_sfc,
                   height=z_sfc,
                   **kwargs,
                   )

    @classmethod
    def from_netcdf(cls, ncfile, timestep=-1, **kwargs):
        """Create a surface model from a netCDF file.

        Parameters:
            ncfile (str): Path to netCDF file.
            timestep (int): Timestep to read (default is last timestep).
        """
        with netCDF4.Dataset(ncfile) as root:
            if 'surface' in root.groups:
                dataset = root['surface']
            else:
                dataset = root

            t = dataset.variables['temperature'][timestep]
            z = dataset.variables['height'][timestep]

        # TODO: Should other variables (e.g. albedo) also be read?
        return cls(temperature=t, height=z, **kwargs)


#TODO: Rename `FixedTemperature`?
class SurfaceFixedTemperature(Surface):
    """Surface model with fixed temperature."""
    def adjust(self, *args, **kwargs):
        """Do not adjust anything for fixed temperature surfaces.

        This function takes an arbitrary number of positional arguments and
        keyword arguments and does nothing.

        Notes:
            Dummy function to fulfill abstract class requirements.
        """
        return


#TODO: Rename `SlabOcean`?
class SurfaceHeatCapacity(Surface):
    """Surface model with adjustable temperature.

    Parameters:
          depth (float): Ocean depth [m].
          **kwargs: Additional keyword arguments are passed to `Surface`.
    """
    def __init__(self, *args, depth=50., **kwargs):
        super().__init__(*args, **kwargs)
        self.rho = constants.density_sea_water
        self.c_p = constants.specific_heat_capacity_sea_water
        self.depth = depth

        self.heat_capacity = self.rho * self.c_p * depth

    def adjust(self, sw_down, sw_up, lw_down, lw_up, timestep):
        """Increase the surface temperature by given heatingrate.

        Parameters:
            sw_down (float): Shortwave downward flux [W / m**2].
            sw_up (float): Shortwave upward flux [W / m**2].
            lw_down (float): Longwave downward flux [W / m**2].
            lw_up (float): Longwave upward flux [W / m**2].
            timestep (float): Timestep in days.
        """
        timestep *= 24 * 60 * 60  # Convert timestep to seconds.

        net_flux = (sw_down - sw_up) + (lw_down - lw_up)

        logger.debug(f'Net flux: {net_flux:.2f} W /m^2')

        self['temperature'] += (timestep * net_flux / self.heat_capacity)

        logger.debug("Surface temperature: {self['temperature'][0]:.4f} K")


class SurfaceHeatSink(SurfaceHeatCapacity):
    """Surface model with adjustable temperature.

    Parameters:
        heat_flux (float):
        **kwargs: Additional keyword arguments are passed to `Surface`.
    """
    def __init__(self, *args, heat_flux=0, **kwargs):
        super().__init__(*args, **kwargs)
        self.heat_flux = heat_flux

    def adjust(self, sw_down, sw_up, lw_down, lw_up, timestep):
        """Increase the surface temperature using given radiative fluxes. Take
        into account a heat sink at the surface, as if heat is transported out
        of the tropics we are modelling.

        Parameters:
            sw_down (float): Shortwave downward flux [W / m**2].
            sw_up (float): Shortwave upward flux [W / m**2].
            lw_down (float): Longwave downward flux [W / m**2].
            lw_up (float): Longwave upward flux [W / m**2].
            timestep (float): Timestep in days.
        """
        timestep *= 24 * 60 * 60  # Convert timestep to seconds.

        net_flux = (sw_down - sw_up) + (lw_down - lw_up)
        sink = self.heat_flux

        logger.debug(f'Net flux: {net_flux:.2f} W /m^2')

        self['temperature'] += (timestep * (net_flux - sink) /
                                self.heat_capacity)

        logger.debug("Surface temperature: {self['temperature'][0]:.4f} K")



# -*- coding: utf-8 -*-
import logging

import typhon
import netCDF4
import numpy as np
from scipy.interpolate import interp1d

from konrad import constants
from konrad import utils
from konrad.component import Component

__all__ = [
    'Atmosphere',
]

logger = logging.getLogger(__name__)


class Atmosphere(Component):
    """Implementation of the atmosphere component."""
    atmosphere_variables = [
        'T',
        'H2O',
        'N2O',
        'O3',
        'O2',
        'CO2',
        'CO',
        'CH4',
        'CFC11',
        'CFC12',
        'CFC22',
        'CCl4',
    ]

    def __init__(self, *args, **kwargs):
        """Atmosphere component. """
        super().__init__(*args, **kwargs)

    def get_default_profile(self, name):
        """Return a profile with default values."""
        try:
            vmr = constants.variable_description[name]['default_vmr']
        except KeyError:
            raise Exception(f'No default specified for "{name}".')
        else:
            return vmr * np.ones(self['plev'].size)

    @classmethod
    def from_atm_fields_compact(cls, atm_fields_compact, **kwargs):
        """Convert an ARTS atm_fields_compact [0] into an atmosphere.

        [0] http://arts.mi.uni-hamburg.de/docserver-trunk/variables/atm_fields_compact

        Parameters:
            atm_fields_compact (typhon.arts.types.GriddedField4):
                Compact set of atmospheric fields.
        """
        def _extract_profile(atmfield, species):
            try:
                arts_key = constants.variable_description[species]['arts_name']
            except KeyError:
                logger.warning(f'No variabel description for "{species}".')
            else:
                return atmfield.get(arts_key, keep_dims=False)

        datadict = {var: _extract_profile(atm_fields_compact, var)
                    for var in cls.atmosphere_variables}
        datadict['plev'] = atm_fields_compact.grids[1]

        return cls.from_dict(datadict, **kwargs)

    @classmethod
    def from_xml(cls, xmlfile, **kwargs):
        """Read atmosphere from XML file containing an ARTS atm_fields_compact.

        Parameters:
            xmlfile (str): Path to XML file.
        """
        # Read the content of given XML file.
        griddedfield = typhon.arts.xml.load(xmlfile)

        # Check if the XML file contains an atm_fields_compact (GriddedField4).
        arts_type = typhon.arts.utils.get_arts_typename(griddedfield)
        if arts_type != 'GriddedField4':
            raise TypeError(
                'XML file contains "{}". Expected "GriddedField4".'.format(
                    arts_type)
            )

        return cls.from_atm_fields_compact(griddedfield, **kwargs)

    @classmethod
    def from_dict(cls, dictionary, **kwargs):
        """Create an atmosphere model from dictionary values.

        Parameters:
            dictionary (dict): Dictionary containing ndarrays.
        """
        # TODO: Currently working for good-natured dictionaries.
        # Consider allowing a more flexibel user interface.

        # Create a Dataset with time and pressure dimension.
        plev = dictionary['plev']
        #TODO: [Discussion] Do we want to read the actual half-level pressure?
        phlev = utils.phlev_from_plev(plev)

        d = cls()
        d.coords = {
            'time': np.array([]),  # time dimension
            'plev': plev,  # pressure level
            'phlev': phlev,  # pressure at halflevels
        }

        for var in cls.atmosphere_variables:
            d.create_variable(var, dictionary.get(var))

        # Calculate the geopotential height.
        d.update_height()

        return d

    @classmethod
    def from_netcdf(cls, ncfile, timestep=-1, **kwargs):
        """Create an atmosphere model from a netCDF file.

        Parameters:
            ncfile (str): Path to netCDF file.
            timestep (int): Timestep to read (default is last timestep).
        """
        def _return_profile(ds, var, ts):
            return (ds[var][ts, :] if 'time' in ds[var].dimensions
                    else ds[var][:])

        with netCDF4.Dataset(ncfile) as root:
            if 'atmosphere' in root.groups:
                dataset = root['atmosphere']
            else:
                dataset = root

            datadict = {var: np.array(_return_profile(dataset, var, timestep))
                        for var in cls.atmosphere_variables
                        if var in dataset.variables
                        }
            datadict['plev'] = np.array(root['plev'][:])

        return cls.from_dict(datadict)

    def to_atm_fields_compact(self):
        """Convert an atmosphere into an ARTS atm_fields_compact."""
        # Store all atmosphere variables including geopotential height.
        variables = self.atmosphere_variables + ['z']

        # Get ARTS variable name from variable description.
        species = [constants.variable_description[var].get('arts_name')
                   for var in variables]

        # Create a GriddedField4.
        atmfield = typhon.arts.types.GriddedField4()

        # Set grids and their names.
        atmfield.gridnames = ['Species', 'Pressure', 'Longitude', 'Latitude']
        atmfield.grids = [
            species, self['plev'], np.array([]), np.array([])
        ]

        # The profiles have to be passed in "stacked" form, as an ndarray of
        # dimensions [species, pressure, lat, lon].
        atmfield.data = np.vstack(
            [self[var].reshape(1, self['plev'].size, 1, 1)
             for var in variables]
        )
        atmfield.dataname = 'Data'

        # Perform a consistency check of the passed grids and data tensor.
        atmfield.check_dimension()

        return atmfield

    def refine_plev(self, pgrid, **kwargs):
        """Refine the pressure grid of an atmosphere object.

        Note:
              This method returns a **new** object,
              the original object is maintained!

        Parameters:
              pgrid (ndarray): New pressure grid [Pa].
            **kwargs: Additional keyword arguments are collected
                and passed to :func:`scipy.interpolate.interp1d`

        Returns:
              Atmosphere: A **new** atmosphere object.
        """
        # Initialize an empty directory to fill it with interpolated data.
        # The dictionary is later used to create a new object using the
        # Atmosphere.from_dict() classmethod. This allows to circumvent the
        # fixed dimension size in xarray.DataArrays.
        datadict = dict()

        datadict['plev'] = pgrid  # Store new pressure grid.

        # Loop over all atmospheric variables...
        for variable in self.atmosphere_variables:
            # and create an interpolation function using the original data.
            f = interp1d(self['plev'], self[variable],
                         axis=-1, fill_value='extrapolate', **kwargs)

            # Store the interpolated new data in the data directory.
            # dims = self.default_dimensions[variable]
            # datadict[variable] = DataArray(f(pgrid), dims=dims)
            datadict[variable] = f(pgrid).ravel()

        # Create a new atmosphere object from the filled data directory.
        # This method also calculates the new phlev coordinates.
        new_atmosphere = type(self).from_dict(datadict)

        # Keep attributes of original atmosphere object.
        # This is **extremely** important because references to e.g. the
        # convection scheme or the humidity handling are stored as attributes!
        new_atmosphere.attrs.update({**self.attrs})

        # Calculate the geopotential height.
        new_atmosphere.update_height()

        return new_atmosphere

    def calculate_height(self):
        """Calculate the geopotential height."""
        g = constants.earth_standard_gravity

        plev = self['plev']  # Air pressure at full-levels.
        phlev = self['phlev']  # Air pressure at half-levels.

        # Air temperature on half levels
        T_phlev = interp1d(plev, self['T'][0, :],
                           fill_value='extrapolate')(phlev)
        # Calculate the air density from current atmospheric state.
        rho_phlev = typhon.physics.density(phlev[:-1], T_phlev[:-1])

        dp = np.hstack((np.array([plev[0] - phlev[0]]), np.diff(plev)))
        # Use the hydrostatic equation to calculate geopotential height from
        # given pressure, density and gravity.
        z = np.cumsum(-dp / (rho_phlev * g))
        return z

    def update_height(self):
        """Update the value for height."""
        z = self.calculate_height()
        # If height is already in Dataset, update its values.
        if 'z' in self.data_vars:
            self.set('z', z)
        # Otherwise create the DataArray.
        else:
            self.create_variable('z', z)

    def get_cold_point_pressure(self):
        """Find the pressure at the cold point.
        The cold point is taken at the coldest temperature below 100 Pa, to
        avoid cold temperatures high in the atmosphere (below about 10 Pa)."""
        p = self['plev']
        T = self['T'][0, :]
        cp = p[np.argmin(T[np.where(p > 100)])]
        return cp

    def get_lapse_rates(self):
        """Calculate the temperature lapse rate at each level."""
        lapse_rate = np.diff(self['T'][0, :]) / np.diff(self['z'][0, :])
        lapse_rate = typhon.math.interpolate_halflevels(lapse_rate)
        lapse_rate = np.append(lapse_rate[0], lapse_rate)
        return np.append(lapse_rate, lapse_rate[-1])

    def get_potential_temperature(self, p0=1000e2):
        """Calculate the potential temperature.

        .. math::
            \theta = T \cdot \left(\frac{p_0}{P}\right)^\frac{2}{7}

        Parameters:
              p0 (float): Pressure at reference level [Pa].

        Returns:
              ndarray: Potential temperature [K].
        """
        # Get view on temperature and pressure arrays.
        T = self['T'][0, :]
        p = self['plev']

        # Calculate the potential temperature.
        return T * (p0 / p) ** (2 / 7)

    def get_static_stability(self):
        """Calculate the static stability.

        .. math::
            \sigma = - \frac{T}{\Theta} \frac{\partial\Theta}{\partial p}

        Returns:
              ndarray: Static stability [K/Pa].
        """
        # Get view on temperature and pressure arrays.
        t = self['T'][0, :]
        p = self['plev']

        # Calculate potential temperature and its vertical derivative.
        theta = self.get_potential_temperature()
        dtheta = np.diff(theta) / np.diff(p)

        return -(t / theta)[:-1] * dtheta

    def get_diabatic_subsidence(self, radiative_cooling):
        """Calculate the diabatic subsidence.

        Parameters:
              radiative_cooling (ndarray): Radiative cooling rates.
                Positive values for heating, negative values for cooling!

        Returns:
            ndarray: Diabatic subsidence [Pa/day].
        """
        sigma = self.get_static_stability()

        return -radiative_cooling[:-1] / sigma

    def get_subsidence_convergence_max(self, radiative_cooling, pmin=10e2):
        """Return index of maximum subsidence convergence.

        Parameters:
            radiative_cooling (ndarray): Radiative cooling rates.
                Positive values for heating, negative values for cooling!
            pmin (float): Lower pressure threshold. The cold point has to
                be below (higher pressure, lower height) that value.

        Returns:
              float: Pressure of maxixum subsidence divergence [Pa].
        """
        plev = self['plev']
        omega = self.get_diabatic_subsidence(radiative_cooling)
        domega = np.diff(omega) / np.diff(plev[:-1])

        # The found maximum is off by 1 due to numerical differentiation in
        # the subsidence calculation. Therefore, return the pressure above.
        max_index = np.argmax(domega[plev[:-2] > pmin]) + 1
        max_plev = plev[max_index]

        self.create_variable('diabatic_convergence_max_index',[max_index])
        self.create_variable('diabatic_convergence_max_plev', [max_plev])

        return max_plev

    def get_coldpoint_plev(self, pmin=10e2):
        """Return the cold point pressure.

        Parameters:
            pmin (float): Minimum pressure threshold. The function does not
                return pressure values smaller than this. This prevents
                finding the upper most level, which is likely to be the
                coldest level.
        """
        T = self['T'][-1, :]
        plev = self['plev'][:]

        return plev[np.argmin(T[plev > pmin])]

    def tracegases_rcemip(self):
        """Set trace gas concentrations according to the RCE-MIP configuration.

        The volume mixing ratios are following the values for the
        RCE-MIP (Wing et al. 2017) and constant throughout the atmosphere.
        """
        concentrations = {
            'CO2': 348e-6,
            'CH4': 1650e-9,
            'N2O': 306e-9,
            'CO': 0,
            'O3': utils.ozone_profile_rcemip(self.get('plev')),
            'CFC11': 0,
            'CFC12': 0,
            'CFC22': 0,
            'CCl4': 0,
        }

        for gas, vmr in concentrations.items():
            self.set(gas, vmr)



import numpy as np
import xarray as xr

from konrad import constants


__all__ = [
    'Component',
]


class Component:
    """Base class for all model components.

    The class implements a light-weight "book-keeping" of all attributes and
    items that are stored in an instance. This allows components to be
    conveniently stored into netCDF files.

    Example implementation of a model component:

    >>> class FancyComponent(Component):
    ...     def __init__(self):
    ...         self.attribute = 'foo'
    ...         self['variable'] = (('dimension',), [42,])
    ...         self.coords = {'dimension': [0]}

    Usage of the implemented auxiliary variables:

    >>> component = FancyComponent()
    >>> component.attrs
    {'attribute': 'foo'}
    >>> component.data_vars
    {'variable': (('dimension',), [42])}

    """
    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        instance._attrs = {}
        instance._data_vars = {}
        instance.coords = {}

        return instance

    def __setattr__(self, name, value):
        object.__setattr__(self, name, value)

        if not name.startswith('_') and name != 'coords':
            self._attrs[name] = value

    def __getattr__(self, name):
        return self._attrs[name]

    @property
    def attrs(self):
        """Dictionary containing all attributes."""
        return self._attrs

    def __setitem__(self, key, value):
        if type(value) is tuple:
            dims, data = value
            self._data_vars[key] = value
        else:
            data = value

        dims = self._data_vars[key][0]
        self._data_vars[key] = (dims, data)

    def __getitem__(self, key):
        if key in self._data_vars:
            return self._data_vars[key][1]
        else:
            return self.coords[key]

    @property
    def data_vars(self):
        """Dictionary containing all data variables and their dimensions."""
        return self._data_vars

    def __repr__(self):
        dims = ', '.join(f'{d}: {np.size(v)}' for d, v in self.coords.items())
        return f'<{self.__class__.__name__}({dims}) object at {id(self)} >'

    @property
    def netcdf_nelem(self):
        """Total number of netCDF elements (attributes and data variables."""
        return len(self.data_vars) + len(self.attrs)

    def to_dataset(self):
        """Convert model component into an `xarray.Dataset`."""
        if self.coords is None:
            raise Exception(
                "Could not create `xarray.Dataset`: `self.coords` not set."
            )
        else:
            self.coords['time'] = [0]
            return xr.Dataset(
                coords=self.coords,
                data_vars=self.data_vars,
                attrs=self.attrs,
            )

    def create_variable(self, name, data=None, dims=None):
        """Create a variable entry in the dataframe."""
        if dims is None:
            dims = constants.variable_description[name].get('dims')

        if data is None:
            data = self.get_default_profile(name)

        ndim = len(dims)
        if ndim == 2 and data.ndim == 1:
            data = data[np.newaxis, :]

        self[name] = (dims, data)

    def set(self, variable, value):
        """Set the values of a variable.

        Parameters:
            variable (str): Variable key.
            value (float or ndarray): Value to assign to the variable.
                If a float is given, all values are filled with it.
        """
        self[variable][:] = value

    def get(self, variable, default=None, keepdims=True):
        """Get values of a given variable.

        Parameters:
            variable (str): Variable key.
            keepdims (bool): If this is set to False, single-dimensions are
                removed. Otherwise dimensions are kept (default).
            default (float): Default value assigned to all pressure levels,
                if the variable is not found.

        Returns:
            ndarray: Array containing the values assigned to the variable.
        """
        try:
            values = self[variable]
        except KeyError:
            if default is not None:
                values = default
            else:
                raise KeyError(f"'{variable}' not found and no default given.")

        return values if keepdims else values.ravel()




import logging
from datetime import datetime

import netCDF4
import numpy as np

from konrad import (constants, __version__)
from konrad.component import Component


__all__ = [
    'NetcdfHandler',
]

logger = logging.getLogger(__name__)


def _move_item_to_index(list, item, index):
    list.insert(index, list.pop(list.index(item)))


def convert_unsupported_types(variable):
    """Convert variables into a netCDF-supported data type."""
    if variable is None:
        return np.nan

    if isinstance(variable, bool):
        return 1 if variable else 0

    if isinstance(variable, str):
        return np.asarray([variable])

    return variable


class NetcdfHandler:
    """A netCDF file handler.

    Usage:
        >>> rce = konrad.RCE(...)
        >>> nc = NetcdfHandler('output.nc', rce)  # create output file
        >>> nc.write(rce)  # write (append) current RCE state to file

    """
    def __init__(self, filename, rce):
        self.filename = filename
        self.rce = rce

        self.udim = 'time'
        self.udim_size = 0
        self.groups = []
        self._component_cache = []

        self.create_file()

    def create_file(self):
        with netCDF4.Dataset(self.filename, mode='w') as root:
            root.setncatts({
                'title': self.rce.experiment,
                'created': datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                'source': f'konrad {__version__}',
                'references': 'https://github.com/atmtools/konrad',
            })

        logger.debug(f'Created "{self.filename}".')

    def create_dimension(self, group, name, data):
        if name not in group.dimensions:
            group.createDimension(name, np.asarray(data).size)

        logger.debug(f'Created dimension "{name}".')

    def create_variable(self, group, name, value, dims=()):
        value = convert_unsupported_types(value)

        variable = group.createVariable(
            varname=name,
            datatype=np.asarray(value).dtype,
            dimensions=dims,
        )
        variable[:] = value

        logger.debug(f'Created variable "{name}".')

        self.append_description(variable)

    def append_description(self, variable):
        desc = constants.variable_description.get(variable.name, {})

        for attribute_name, value in desc.items():
            logger.debug(
                f'Added attribute "{attribute_name}" to "{variable.name}".')
            setattr(variable, attribute_name, value)

    def create_group(self, component, groupname):
        with netCDF4.Dataset(self.filename, 'a') as root:
            group = root.createGroup(groupname)

            for attr, value in component.attrs.items():
                self.create_variable(group, attr, value)

            for name, coord in component.coords.items():
                self.create_dimension(root, name, coord)
                if name not in root.variables:
                    self.create_variable(root, name, coord, (name,))

            for varname, (dims, data) in component.data_vars.items():
                if varname not in group.variables:
                    self.create_variable(group, varname, data, dims)

            logger.debug(f'Created group "{groupname}".')

            self.groups.append(groupname)

    def append_group(self, component, groupname):
        with netCDF4.Dataset(self.filename, 'a') as root:
            group = root.groups[groupname]

            for varname, (dims, data) in component.data_vars.items():
                if self.udim not in dims:
                    continue

                s = [self.udim_size if dim == self.udim else slice(None)
                     for dim in dims]

                group.variables[varname][tuple(s)] = data

    def expand_unlimitied_dimension(self):
        with netCDF4.Dataset(self.filename, 'a') as root:
            self.udim_size = root.dimensions[self.udim].size

            root[self.udim][self.udim_size] = self.rce.get_hours_passed()

    def get_components(self):
        """Return a list of non-empty non-private model components."""
        if len(self._component_cache) == 0:
            for attr in dir(self.rce):
                if ((not attr.startswith('_')
                     and isinstance(getattr(self.rce, attr), Component)
                     and getattr(self.rce, attr).netcdf_nelem > 0)):
                    self._component_cache.append(attr)

        # Ensure that the atmosphere component is stored first as it holds
        # the common coordinates `plev` and `phlev`.
        self._component_cache.sort()
        _move_item_to_index(self._component_cache, 'atmosphere', 0)

        logger.debug(f'Components for netCDF file: "{self._component_cache}".')

        return self._component_cache

    def initialize_file(self):
        for component in self.get_components():
            self.create_group(getattr(self.rce, component), component)

        with netCDF4.Dataset(self.filename, 'a') as root:
            root.variables[self.udim][:] = 0

    def append_to_file(self):
        self.expand_unlimitied_dimension()
        for component in self.get_components():
            self.append_group(getattr(self.rce, component), component)

    def write(self):
        """Write current state of the RCE model to the netCDF file."""
        if len(self.groups) == 0:
            self.initialize_file()
        else:
            self.append_to_file()



"""Define the abstract Radiation base class. """
import abc
import logging

import numpy as np
import xarray as xr
from scipy.interpolate import interp1d

from konrad import constants
from konrad.component import Component
from konrad.utils import append_description


logger = logging.getLogger(__name__)

__all__ = [
    'Radiation',
]

# Subclasses of `Radiation` need to define a `Radiation.calc_radiation()`
# method that returns a `xr.Dataset` containing at least following variables:
REQUIRED_VARIABLES = [
    'net_htngrt',
    'lw_flxd',
    'lw_flxu',
    'sw_flxd',
    'sw_flxu',
]


class Radiation(Component, metaclass=abc.ABCMeta):
    """Abstract base class to define requirements for radiation models."""
    def __init__(self, zenith_angle=47.88, diurnal_cycle=False, bias=None):
        """Return a radiation model.

        Parameters:
            zenith_angle (float): Zenith angle of the sun.
                The default angle of 47.88 degree results in 342 W/m^2
                solar insolation at the top of the atmosphere when used
                together with a solar constant of 510 W/m^2.
            diurnal_cycle (bool): Toggle diurnal cycle of solar angle.
            bias (dict-like): A dict-like object that stores bias
                corrections for the diagnostic variable specified by its key,
                e.g. `bias = {'net_htngrt': 2}`.
        """
        super().__init__()

        self.zenith_angle = zenith_angle
        self.diurnal_cycle = diurnal_cycle

        self.current_solar_angle = 0

        self._bias = bias

        self['lw_htngrt'] = (('time', 'plev'), None)
        self['lw_htngrt_clr'] = (('time', 'plev'), None)
        self['lw_flxu'] = (('time', 'phlev'), None)
        self['lw_flxd'] = (('time', 'phlev'), None)
        self['lw_flxu_clr'] = (('time', 'phlev'), None)
        self['lw_flxd_clr'] = (('time', 'phlev'), None)
        self['sw_htngrt'] = (('time', 'plev'), None)
        self['sw_htngrt_clr'] = (('time', 'plev'), None)
        self['sw_flxu'] = (('time', 'phlev'), None)
        self['sw_flxd'] = (('time', 'phlev'), None)
        self['sw_flxu_clr'] = (('time', 'phlev'), None)
        self['sw_flxd_clr'] = (('time', 'phlev'), None)

        self['net_htngrt'] = (('time', 'plev'), None)
        self['toa'] = (('time',), None)

    @abc.abstractmethod
    def calc_radiation(self, atmosphere, surface, cloud):
        pass

    def update_heatingrates(self, atmosphere, surface, cloud):
        """Returns `xr.Dataset` containing radiative transfer results."""
        self.calc_radiation(atmosphere, surface, cloud)

        # self.correct_bias(rad_dataset)

        self.derive_diagnostics()

    @staticmethod
    def check_dataset(dataset):
        """Check if a given dataset contains all required variables."""
        for key in REQUIRED_VARIABLES:
            if key not in dataset.variables:
                raise KeyError(
                    f'"{key}" not present in radiative transfer results.'
                )

    def correct_bias(self, dataset):
        """Apply bias correction."""
        # Interpolate biases passed as `xr.Dataset`.
        if isinstance(self._bias, xr.Dataset):
            bias_dict = {}
            for key in self._bias.data_vars:
                zdim = self._bias[key].dims[0]
                x = self._bias[zdim].values
                y = self._bias[key].values

                f_interp = interp1d(x, y, fill_value='extrapolate')
                bias_dict[key] = f_interp(dataset[zdim].values)

            self._bias = bias_dict

        if self._bias is not None:
            for key, value in self._bias.items():
                if key not in dataset.indexes:
                    dataset[key] -= value


    def derive_diagnostics(self):
        """Derive diagnostic variables from radiative transfer results."""
        # Net heating rate.
        self['net_htngrt'] = self['lw_htngrt'] + self['sw_htngrt']

        # Radiation budget at top of the atmosphere (TOA).
        self['toa'] = (
            (self['sw_flxd'][:, -1] + self['lw_flxd'][:, -1]) -
            (self['sw_flxu'][:, -1] + self['lw_flxu'][:, -1])
        )

    @staticmethod
    def heatingrates_from_fluxes(pressure, downward_flux, upward_flux):
        """Calculate heating rates from radiative fluxes.

        Parameters:
            pressure (ndarray): Pressure half-levels [Pa].
            downward_flux (ndarray): Downward radiative flux [W/m^2].
            upward_flux (ndarray): Upward radiative flux [W/m^2].

        Returns:
            ndarray: Radiative heating rate [K/day].
        """
        c_p = constants.isobaric_mass_heat_capacity
        g = constants.earth_standard_gravity

        q = g / c_p * np.diff(upward_flux - downward_flux) / np.diff(pressure)
        q *= 3600 * 24

        return q

    def adjust_solar_angle(self, time):
        """Adjust the zenith angle of the sun according to time of day.

        Parameters:
            time (float): Current time [days].
        """
        # When the diurnal cycle is disabled, use the constant zenith angle.
        if not self.diurnal_cycle:
            self.current_solar_angle = self.zenith_angle
            return

        # The local zenith angle, calculated from the latitude and longitude.
        # Seasons are not considered.
        solar_angle = np.rad2deg(np.arccos(
            np.cos(np.deg2rad(self.zenith_angle)) * 
            np.cos(2 * np.pi * time)))

        self.current_solar_angle = solar_angle



# -*- coding: utf-8 -*-
"""Implementation of a radiative-convective equilibrium model (RCE).
"""
import logging
from numbers import Number

import numpy as np

from konrad import utils
from konrad import netcdf
from konrad.radiation import RRTMG
from konrad.ozone import (Ozone, OzonePressure)
from konrad.humidity import (Humidity, FixedRH)
from konrad.surface import (Surface, SurfaceHeatCapacity)
from konrad.cloud import (Cloud, ClearSky)
from konrad.convection import (Convection, HardAdjustment, RelaxedAdjustment)
from konrad.lapserate import (LapseRate, MoistLapseRate)
from konrad.upwelling import (Upwelling, NoUpwelling)

logger = logging.getLogger(__name__)

__all__ = [
    'RCE',
]


class RCE:
    """Interface to control the radiative-convective equilibrium simulation.

    Examples:
        Create an object to setup and run a simulation:
        >>> import konrad
        >>> rce = konrad.RCE(...)
        >>> rce.run()
    """

    def __init__(self, atmosphere, timestep='3h', max_duration='5000d',
                 outfile=None, experiment='RCE', writeevery='1d', delta=1e-4,
                 radiation=None, ozone=None, humidity=None, surface=None,
                 cloud=None, convection=None, lapserate=None, upwelling=None):
        """Set-up a radiative-convective model.

        Parameters:
            atmosphere (Atmosphere): `konrad.atmosphere.Atmosphere`.
            timestep (float or str): Model time step (per iteration).
                If float, time step shall be given in days.
                If str, a timedelta string may be given
                (see :func:`konrad.utils.parse_fraction_of_day`).
            max_duration (float or str): Maximum duration of the simulation.
                The duration is given in model time:
                    If float, maximum duration in days.
                    If str, a timedelta string
                    (see `konrad.utils.parse_fraction_of_day`).
            outfile (str): netCDF4 file to store output.
            experiment (str): Experiment description (stored in netCDF output).
            writeevery(int, float or str): Set output frequency.
                Values can be given in:
                    int: Every nth iteration
                    float: Every nth day in model time
                    str: a timedelta string
                    (see `konrad.utils.parse_fraction_of_day`).
            delta (float): Stop criterion. If the heating rate is below this
                threshold for all levels, skip further iterations. Values
                are given in K/day.
            radiation (konrad.radiation): Radiation model.
                Defaults to :class:`konrad.radiation.RRTMG`.
            ozone (konrad.ozone): Ozone model.
                Defaults to :class:`konrad.ozone.OzonePressure`.
            humidity (konrad.humidity): Humidity model.
                Defaults to :class:`konrad.humidity.FixedRH`.
            surface (konrad.surface): Surface model.
                Defaults to :class:`konrad.surface.SurfaceHeatCapacity`.
            cloud (konrad.cloud): Cloud model.
                Defaults to :class:`konrad.cloud.ClearSky`.
            convection (konrad.humidity.Convection): Convection scheme.
                Defaults to :class:`konrad.convection.HardAdjustment`.
            lapserate (konrad.lapse.LapseRate): Lapse rate handler.
                Defaults to :class:`konrad.lapserate.MoistLapseRate`.
            upwelling (konrad.upwelling.Upwelling):
                Defaults to :class:`konrad.upwelling.NoUpwelling`.
        """
        # Sub-models.
        self.atmosphere = atmosphere
        if radiation is None:
            self.radiation = RRTMG()
        else:
            self.radiation = radiation

        self.ozone = utils.return_if_type(ozone, 'ozone',
                                          Ozone, OzonePressure())

        self.humidity = utils.return_if_type(humidity, 'humidity',
                                             Humidity, FixedRH())
        self.surface = utils.return_if_type(surface, 'surface',
                                            Surface, SurfaceHeatCapacity())
        self.cloud = utils.return_if_type(cloud, 'cloud',
                                          Cloud,
                                          ClearSky(self.atmosphere.get(
                                              'z', keepdims=False))
                                          )
        self.convection = utils.return_if_type(convection, 'convection',
                                               Convection, HardAdjustment())

        self.lapserate = utils.return_if_type(lapserate, 'lapserate',
                                              LapseRate, MoistLapseRate())

        self.upwelling = utils.return_if_type(upwelling, 'upwelling',
                                              Upwelling, NoUpwelling())

        self.max_duration = utils.parse_fraction_of_day(max_duration)
        self.timestep = utils.parse_fraction_of_day(timestep)
        self.writeevery = utils.parse_fraction_of_day(writeevery)

        self.max_iterations = np.ceil(self.max_duration / self.timestep)
        self.niter = 0

        self.delta = delta
        self.deltaT = None
        self.converged = False

        self.outfile = outfile
        self.nchandler = None
        self.experiment = experiment

        logging.info('Created Konrad object:\n{}'.format(self))

    def __repr__(self):
        retstr = '{}(\n'.format(self.__class__.__name__)
        # Loop over all public object attributes.
        for a in filter(lambda k: not k.startswith('_'), self.__dict__):
            retstr += '    {}={},\n'.format(a, getattr(self, a))
        retstr += ')'

        return retstr

    def get_hours_passed(self):
        """Return the number of hours passed since model start.

        Returns:
            float: Hours passed since model start.
        """
        return self.niter * 24 * self.timestep

    def is_converged(self):
        """Check if the atmosphere is in radiative-convective equilibrium.

        Returns:
            bool: ``True`` if converged, else ``False``.
        """
        # TODO: Implement proper convergence criterion (e.g. include TOA).
        return np.all(np.abs(self.deltaT) < self.delta)

    def check_if_write(self):
        """Check if current timestep should be appended to output netCDF.

        Do not write, if no output file is specified.

        Returns:
            bool: True, if timestep should be written.
        """
        if self.outfile is None:
            return False

        if isinstance(self.writeevery, int):
            return self.niter % self.writeevery == 0
        elif isinstance(self.writeevery, float):
            # Add `0.5 * dt` to current timestep to make float comparison more
            # robust. Otherwise `3.3 % 3 < 0.3` is True.
            r = (((self.niter + 0.5) * self.timestep) % self.writeevery)
            return r < self.timestep

    def run(self):
        """Run the radiative-convective equilibrium model."""
        logger.info('Start RCE model run.')

        # Initialize surface pressure to be equal to lowest half-level
        # pressure. This is consistent with handling in PSrad.
        self.surface.pressure = self.atmosphere['phlev'][0]

        # Main loop to control all model iterations until maximum number is
        # reached or a given stop criterion is fulfilled.
        while self.niter < self.max_iterations:
            if self.niter % 100 == 0:
                # Write every 100th time step in loglevel INFO.
                logger.info(f'Enter iteration {self.niter}.')
            else:
                # All other iterations are only logged in DEBUG level.
                logger.debug(f'Enter iteration {self.niter}.')

            self.radiation.adjust_solar_angle(self.get_hours_passed() / 24)
            self.radiation.update_heatingrates(
                atmosphere=self.atmosphere,
                surface=self.surface,
                cloud=self.cloud,
            )

            # Apply heatingrates/fluxes to the the surface.
            self.surface.adjust(
                sw_down=self.radiation['sw_flxd'][0, 0],
                sw_up=self.radiation['sw_flxu'][0, 0],
                lw_down=self.radiation['lw_flxd'][0, 0],
                lw_up=self.radiation['lw_flxu'][0, 0],
                timestep=self.timestep,
            )

            # Save the old temperature profile. They are compared with
            # adjusted values to check if the model has converged.
            T = self.atmosphere['T'].copy()

            # Caculate critical lapse rate.
            critical_lapserate = self.lapserate(self.atmosphere)

            # Apply heatingrates to temperature profile.
            self.atmosphere['T'] += (self.radiation['net_htngrt'] *
                                     self.timestep)

            # Convective adjustment
            self.convection.stabilize(
                atmosphere=self.atmosphere,
                lapse=critical_lapserate,
                timestep=self.timestep,
                surface=self.surface,
            )

            # Upwelling induced cooling
            self.upwelling.cool(
                atmosphere=self.atmosphere,
                radheat=self.radiation['net_htngrt'][0, :],
                timestep=self.timestep,
            )

            # TODO: Consider implementing an Atmosphere.update_diagnostics()
            #  method to include e.g. convective top in the output.
            self.atmosphere.update_height()
            z = self.atmosphere.get('z')[0, :]
            if isinstance(self.convection, HardAdjustment) or isinstance(
                    self.convection, RelaxedAdjustment):
                self.convection.calculate_convective_top_height(z)

            # Update the ozone profile.
            self.ozone(
                atmosphere=self.atmosphere,
                convection=self.convection,
                timestep=self.timestep,
                zenith=self.radiation.current_solar_angle
            )

            # Update the humidity profile.
            self.atmosphere['H2O'][0, :] = self.humidity(
                self.atmosphere,
                convection=self.convection,
                surface=self.surface,
            )

            self.cloud.update_cloud_profile(self.atmosphere,
                                            convection=self.convection)

            # Calculate temperature change for convergence check.
            self.deltaT = (self.atmosphere['T'] - T) / self.timestep

            # Check, if the current iteration is scheduled to be written.
            if self.check_if_write():
                if self.nchandler is None:
                    self.nchandler = netcdf.NetcdfHandler(
                        filename=self.outfile, rce=self)

                self.nchandler.write()

            # Check if the model run has converged to an equilibrium state.
            if self.is_converged():
                # If the model is converged, skip further iterations. Success!
                logger.info(f'Converged after {self.niter} iterations.')
                break
            # Otherweise increase the iteration count and go on.
            else:
                self.niter += 1
        else:
            logger.info('Stopped after maximum number of iterations.')
